/****** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is XpcomViewer code.
 * 
 * The Initial Developer of the Original Code is NeutronStar, s.r.o.
 * Portions created by NeutronStar, s.r.o. are Copyright (C) 2005 
 * NeutronStar, s.r.o.. All Rights Reserved.
 * 
 * Contributor(s):
 *  - Ondrej Donek, <ondrejd@gmail.com>, original author
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the LGPL or the GPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 ****** END LICENSE BLOCK *****/



/**
 * Main gXpcomViewer class.
 */
var gXpcomViewer = 
{
    /** 
     * Used XPCOM comonents 
     */
   	nsIRDFResource 	: Components.interfaces.nsIRDFResource,
	nsIRDFNode 		: Components.interfaces.nsIRDFNode,
	nsIRDFLiteral 	: Components.interfaces.nsIRDFLiteral,
	nsIRDFDS		: Components.interfaces.nsIRDFDataSource,
	nsILocalFile	: Components.interfaces.nsILocalFile,
	inMemoryDS 		: "@mozilla.org/rdf/datasource;1?name=in-memory-datasource",
	
	prefSrv 		: Components.classes["@mozilla.org/preferences-service;1"].
					  getService (Components.interfaces.nsIPrefService),
	rdfSrv 			: Components.classes["@mozilla.org/rdf/rdf-service;1"].
    				  getService (Components.interfaces.nsIRDFService),
	IOSrv			: Components.classes["@mozilla.org/network/io-service;1"].
					  getService (Components.interfaces.nsIIOService),
    
	
	/** 
	 * Instance of nsIPrefBranch 
	 */
	get mPrefs () { return this.prefSrv.getBranch ("extensions.xpcomviewer."); },
	
	
	/** 
	 * Properties describing our rdf datasource 
	 */
	mDataSource 	: null,
	mDataSourceFile : null,
	// Here are some getters for variables we oftenly use when working 
	// with our rdf datasource.
	get namespace () { return "http://www.mozilla.org/NC-xpcom#"; },
	get uriBase () { return "rdf:xpcom"; },
	get uriClasses () { return "rdf:xpcom:classes"; },
	get uriIfaces () { return "rdf:xpcom:interfaces"; },
	get uriResults () { return "rdf:xpcom:results"; },
	get uriBookmarks () { return "rdf:xpcom:bookmarks"; },
	get uriNotes () { return "rdf:xpcom:notes"; },
    // resources 
	get resBase () { return this.rdfSrv.GetResource (this.uriBase); },
	get resClasses () { return this.rdfSrv.GetResource (this.uriClasses); },
	get resIfaces () { return this.rdfSrv.GetResource (this.uriIfaces); },
	get resResults () { return this.rdfSrv.GetResource (this.uriResults); },
	get resBookmars () { return this.rdfSrv.GetResource (this.uriBookmarks); },
	get resNotes ()	{ return this.rdfSrv.GetResource (this.uriNotes); },
    // arcs
    get arcLabel () { return this.rdfSrv.GetResource (this.namespace + "label"); },
    get arcValue () { return this.rdfSrv.GetResource (this.namespace + "value"); },
    get arcType () { return this.rdfSrv.GetResource (this.namespace + "type"); },
    get arcShow () { return this.rdfSrv.GetResource (this.namespace + "show"); },
    get arcLoaded () { return this.rdfSrv.GetResource (this.namespace + "loaded"); },
	get arcClass ()	{ return this.rdfSrv.GetResource (this.namespace + "class"); },
    get arcContractID () { return this.rdfSrv.GetResource (this.namespace + "contractid"); },
	get arcIface () { return this.rdfSrv.GetResource (this.namespace + "interface"); },
    get arcClassID () { return this.rdfSrv.GetResource (this.namespace + "clsid"); },
    get predNote () { return this.rdfSrv.GetResource (this.namespace + "note"); },
    get arcHasNote () { return this.rdfSrv.GetResource (this.namespace + "hasNote"); },
    get arcBookmark () { return this.rdfSrv.GetResource (this.namespace + "bookmark"); },
    get arcHasBookmark () { return this.rdfSrv.GetResource (this.namespace + "hasBookmark"); },
    // literals
    get litTrue () { return this.rdfSrv.GetLiteral ("true"); },
    /*get litFalse () { return this.rdfSrv.GetLiteral ("false"); },*/
    get litIsNote () { return this.rdfSrv.GetLiteral ("isNote"); },
    get litNoNote () { return this.rdfSrv.GetLiteral ("noNote"); },
    get litIsBookmark () { return this.rdfSrv.GetLiteral ("isBookmark"); },
    get litNoBookmark () { return this.rdfSrv.GetLiteral ("noBookmark"); },
    get litTypeClass () { return this.rdfSrv.GetLiteral ("class"); },
    get litTypeIface () { return this.rdfSrv.GetLiteral ("interface"); },
    get litTypeResult () { return this.rdfSrv.GetLiteral ("result"); },
    // getter for root container
    rootCont : null,
    /* get rootCont () {
        try {
            return Components.classes["@mozilla.org/rdf/container;1"].
                createInstance (Components.interfaces.nsIRDFContainer).
                Init (this.mDataSource, this.resBase);
        } catch (e) {
            return Components.classes["@mozilla.org/rdf/container-utils;1"].
                   createInstance (Components.interfaces.nsIRDFContainerUtils).
                   MakeSeq (this.mDataSource, this.resBase)
        }
    }, */
    
	
	/** 
	 * Getters for pointers to UI elements.
	 */
	get mStrBundle () { return document.getElementById ("stringBundle"); },
	get mBrandStrBundle () { return document.getElementById ("brandStringBundle"); },
	// main XpcomViewer tree
	get mTree () { return document.getElementById ("tree"); },
	
    
	/** 
     * Internal property containg uri which is currently used as value 
     * of ref attribute on main tree.
     */
	_lastView : null,
    
    /**
     * Flag for OnUnload method - if main datasource have to be saved.
     */
    __TO_BE_SAVED : false,
    
    /**
     * Internal property containg boolean value if is currently any filter 
     * applied on datasource or not.
     * It also switch disable attribute on removeFilter-broadcaster immediately 
     * when is setter called.
     */
    __isFilterApplied : false,
    set _isFilterApplied (aApplied) {
        var broadcaster = document.getElementById ("removeFilter-broadcaster");
        
        if (!aApplied){
            broadcaster.setAttribute ("disabled", true);
        } else {
            if (broadcaster.hasAttribute ("disabled"))
                broadcaster.removeAttribute ("disabled");
        }
        this.__isFilterApplied = aApplied;
    },
    get _isFilterApplied () { 
        return this.__isFilterApplied; 
    },
	
	
	/////////////////////////////////////////////
	// Events-processing methods
	    
    
	/**
	 * When XPCOMView application window started.
	 */
	onLoad : function ()
	{
        window.focus ();
		
		// initialize datasource file (which is stored in user profile 
		// folder)
		this.initializeDataSourceFile ();
		
		// Initialize last used view.
		if (this.mPrefs.prefHasUserValue ("workspace.lastview"))
		    this._lastView = this.mPrefs.getCharPref ("workspace.lastview");
		else
		    this._lastView = this.uriClasses;
		
		// if file already exist load it
		if (this.mDataSourceFile.exists ())
		{
		    this.loadDataSourceFromFile ();
		}
		else
		{
		    // otherwise create new blank datasource
		    this.createBlankDataSource ();
		    this.saveDataSourceToFile ();
		}
            
		// refresh main tree
		if (this.mDataSource instanceof this.nsIRDFDS)
		    this.refreshTree (this._lastView);
		else
		{
		    /**
		     * Datasource wasn't created so we can expected critical 
		     * error ;) -> display this state to user and offer him 
		     * close app.
		     */
		    if (confirm (this.mStrBundle ("confirmQuitDueError")))
		        window.close ();
		}
        
        // initalize bookmarks datasource
        gBookmarksDataSource.init ([document.getElementById ("allBookmarks-menu")]);
        // initalize notes datasource
        gNotesDataSource.init ([document.getElementById ("allNotes-menu")]);
	},
	

	/**
	 * On gXpcomViewer unload. This function only can finish pendling 
	 * operations but can not stop closing application!
	 */
	onUnLoad : function ()
	{
		// if any change occured save main xpcom datasource
        if (this.__TO_BE_SAVED__)
            this.saveDataSourceToFile ();
        
        // save bookmarks datasource
        gBookmarksDataSource.saveDataSourceToFile ();
        // save notes datasource
        gNotesDataSource.saveDataSourceToFile ();
        
        this.mPrefs.setCharPref ("workspace.lastview", this._lastView);
	},
	
	
	/**
	 * When tree item is selected.
	 */
	onTreeItemSelected : function ()
	{
		var resources = this.getSelectedTreeItems ();
		
		for (var res in resources)
		{
			var cmpType = "";
			var target = this.mDataSource.GetTarget (resources[res], 
			             this.arcType, true);
				
			if (target instanceof this.nsIRDFLiteral)
				cmpType = target.Value;
				
			if (!this.mDataSource.hasArcOut (resources[res], this.arcLoaded))
			{
				var reload = false;
				
				// try to load scope of selected object. If loading is 
				// successfull refresh also main xpcom tree
				switch (cmpType)
				{
				    case "class": 
					    reload = this.loadClassAtResource (resources[res]);
					    break;
					case "interface":
					    reload = this.loadInterfaceAtResource (resources[res]);
					    break;
				}
				
				if (reload) this.refreshTree (this._lastView);
			}
		}
	}, // end gXpcomViewer.onTreeItemSelected ()
	
	
	/**
	 * Fired when user clicked for show customize toolbar popup menu. 
	 * This method enumerate all toolbars under the main toolbox element 
	 * and create appropriate menuitems (checkbox type).
	 * Deleting of these newly generated menuitems is performed by 
	 * gXpcomViewer.onCustomizeToolbarsMenupopupHide ().
	 */
	onCustomizeToolbarsMenupopupShow : function ()
	{
	   var xul_ns = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";
	   var popup = document.getElementById ("toolbox-context-menu");
	   var popupSep = document.getElementById ("enumeratedToolbarsSeparator");
	   var toolbox = document.getElementById ("xpcomviewer-toolbox");
	   var toolbars = toolbox.getElementsByTagName ("toolbar");
	   
	   for (var toolbar in toolbars)
	   {
	       if (toolbar.hasAttribute ("customizable") && 
	           toolbar.hasAttribute ("toolbarname"))
	       {
	           var menuitem = document.createElementNS(XUL_NS, "menuitem");
	           menuitem.setAttribute ("type", "checkbox");
	           menuitem.setAttribute ("label", toolbar.getAttribute ("toolbarname"));
	           menuitem.setAttribute ("oncommand", "gXpcomViewer.setUpToolbar (this.label);");
	           
	           if (!toolbar.hasAttribute ("collapsed"))
	               menuitem.setAttribute ("checked", true);
	       }
	   }
	}, // end gXpcomViewer.onCustomizeToolbarsMenupopupShow ()
	
	
	/** 
	 * Fired when is toolbars-context-menu hiding. This method 
	 * delete newly created menuitems (in onCustomizeToolbarsMenupopupShow ()
	 * method).
	 */
	onCustomizeToolbarsMenupopupHide : function ()
	{
	   var popup = document.getElementById ("toolbox-context-menu");
	   var menuitems = popup.getElementsByTagName ("menuitem");
	   var itemsToDelete = [];
	   
	   for (var menuitem in menuitems)
	       if (menuitem.hasAttribute ("type"))
	           if (menuitem.getAttribute ("type") == "checkbox")
	               itemsToDelete.push (menuitem);
	   
	   for (var item in itemsToDelete)
	       popup.removeChild (item);
	}, // end gXpcomViewer.onCustomizeToolbarsMenupopupHide ()
	
    
	/**
	 * This method is called from generated menuitems in toolbars-context-menu. 
	 * It toggle collapsed attribute on toolbar with given toolbarname.
	 * 
	 * @param  string  Toolbarname
	 */
	setUpToolbar : function (aToolbarName)
	{
	   var toolbox = document.getElementById ("xpcomviewer-toolbox");
	   var toolbars = toolbox.getElementsByTagName ("toolbar");
	   
	   for (var toolbar in toolbars)
	   {
	       if (toolbar.hasAttribute ("customizable") && 
	           toolbar.hasAttribute ("toolbarname"))
	       {
	           if (toolbar.getAttribute ("toolbarname") == aToolbarName)
	           {
	               if (toolbar.hasAttribute ("collapsed"))
	                   toolbar.removeAttribute ("collapsed");
	               else
	                   toolbar.setAttribute ("collapsed", true);
	           }
	       }
	   }
	}, // end gXpcomViewer.setUpToolbar ()
	
	
	
	/////////////////////////////////////////////
	/// Common commands methods
		
	/** Show about dialog. */
	showAboutDialog : function () {
		window.openDialog ("chrome://xpcomviewer/content/about.xul", 
		                   "xpcomviewer:about", 
		                   "chrome,extrachrome,dialog,modal,centerscreen");
	},
	/** Show preferences dialog. */
	showPreferences : function () {
		window.openDialog ("chrome://xpcomviewer/content/preferences.xul", 
		                   "xpcomviewer:about", "chrome,all");
	},
	/** Opens toolbar customization dialog. */
	customizeToolbars : function () {
		window.openDialog ("chrome://global/content/customizeToolbar.xul", 
						   "CustomizeToolbar", "chrome,all,dependent", 
						   document.getElementById ("xpcomviewer-toolbox"));
	},
	/** Show bookmark manager dialog. */
	showBookmarksManager : function () {
		window.openDialog ("chrome://xpcomviewer/content/bookmarksmanager.xul", 
		                   "xpcomviewer:bookmarksmanager", "chrome,all,dependent");
	},
	/** Show notes manager dialog. */
	showNotesManager : function () {
		window.openDialog ("chrome://xpcomviewer/content/notesmanager.xul", 
		                   "xpcomviewer:notesmanager", "chrome,all,dependent");
	},
		
	
	
	/////////////////////////////////////////////
	/// Main tree methods and methods related to 
	/// main xpcom datasource
	
	
	/** 
	 * Refresh tree with given reference uri. 
	 * 
	 * @param  string
	 */
	refreshTree : function (aUri)
	{
	    // check if container with given uri exists
       
        try {
            var cont = Components.classes["@mozilla.org/rdf/container;1"].
                createInstance (Components.interfaces.nsIRDFContainer).
                Init (this.mDataSource, this.rdfSrv.GetResource (aUri));
        }
        catch (e)
        {
            // if no create it
            if (aUri == this.uriClasses)
                this.createComponentsClasses ();
            else if (aUri == this.uriIfaces)
                this.createComponentsInterfaces ();
            else if (aUri == this.uriResults)
                this.createComponentsResults ();
        }
       
        // set up last view and remove current datasources
        this._lastView = aUri;
	    this.removeTreeDataSources (this.mTree);
       
        // set new view
	    this.mTree.setAttribute ("ref", aUri);
	    this.mTree.database.AddDataSource (this.mDataSource);
        this.mTree.builder.rebuild ();
	}, // end gXpcomViewer.refreshTree (aUri)
	
	
	/**
	 * Removes all attached rdf datasources of given tree.
	 * 
	 * @param	object
	 */
	removeTreeDataSources : function (aTree)
	{
		var datasources = aTree.database.GetDataSources ();
		
		while (datasources.hasMoreElements ())
		{
			var ds = datasources.getNext ();
			try
			{
				aTree.database.RemoveDataSource (ds);
			}
			catch (e)
			{ }
		}
	}, // end gXpcomViewer.removeTreeDataSources (aTree)
	
	
	/**
	 * Returns array of resources (nsIRDFResource) of selected tree items.
	 * 
	 * @return 	object	Array of resources
	 */
	getSelectedTreeItems : function ()
	{
		var start = new Object ();
		var end = new Object ();
		var numRanges = this.mTree.view.selection.getRangeCount ();
		var resources = [];
		
		for (var t=0; t<numRanges; t++)
		{
			this.mTree.view.selection.getRangeAt (t, start, end);
			
			for (var v=start.value; v<=end.value; v++)
				resources.push (this.mTree.view.getResourceAtIndex (v));
				// resources.push (this.rdfSrv.GetResource (this.mTree.view.GetItemAtIndex (v).id));
		}
		
		return resources;
	}, // end gXpcomViewer.getSelectedTreeItems ()
	
	
	/**
	 * Get details from component on given resource and put 
	 * these details into properties tree.
	 * 
	 * @param  nsIRDFResource
	 */
	loadClassAtResource : function (aRes)
	{
	    var res = this.rdfSrv.GetResource (aRes.Value);
	    var target = this.mDataSource.GetTarget (aRes, this.arcLabel, true);
	    var clName = "";
	   
	    if (target instanceof this.nsIRDFLiteral)
	        clName = target.Value;
	   
	    if (clName != "")
	    {
	        var clCont = Components.classes["@mozilla.org/rdf/container-utils;1"].
                         createInstance (Components.interfaces.nsIRDFContainerUtils).
                         MakeSeq (this.mDataSource, res);
	       
	        for (var iface in Components.interfaces)
	        {
                // is interface supported ?
                var queryIface = false; 
                try {
                    Components.classes[clName].createInstance (Components.interfaces.nsISupports).
                        QueryInterface (Components.interfaces[iface], {});
                    queryIface = true;
                } catch (e) {
                    queryIface = false;
                }
                
                // create instance
	            var cc = null;
                if (queryIface)
                {
                    // Some interfaces have to be initialized by getService 
                    // instead of createInstance. These interfaces have oftenly 
                    // words 'service' or 'manager' contained in their title...
                    
                    var ifaceName = new String(iface).toLowerCase ();
                    
                    if (ifaceName.indexOf ("manager") != -1 || 
                        ifaceName.indexOf ("service") != -1) {
                        cc = Components.classes[clName].
                             getService (Components.interfaces[iface]); 
                    } else {
                        cc = Components.classes[clName].
                             createInstance (Components.interfaces[iface]);
                    }
                }
	           
	            if (cc != null)
	            {
	                var ifaceUri = aRes.Value + ":" + iface;
	                var ifaceRes = this.rdfSrv.GetResource (ifaceUri);
	                this.mDataSource.Assert (ifaceRes, this.arcLabel, this.rdfSrv.GetLiteral (iface), true);
	                this.mDataSource.Assert (ifaceRes, this.arcType, this.litTypeIface, true);
                    this.mDataSource.Assert (ifaceRes, this.arcShow, this.litTrue, true);
                    this.mDataSource.Assert (ifaceRes, this.arcHasNote, this.litNoNote, true);
                    this.mDataSource.Assert (ifaceRes, this.arcHasBookmark, this.litNoBookmark, true);
                    
	                var ifaceCont = Components.classes["@mozilla.org/rdf/container-utils;1"].
                                    createInstance (Components.interfaces.nsIRDFContainerUtils).
                                    MakeSeq (this.mDataSource, ifaceRes);
	               
	                for (var prop in cc)
	                {
	                    var propRes = this.rdfSrv.GetResource (ifaceUri + ":" + prop);
	                    this.mDataSource.Assert (propRes, this.arcLabel, this.rdfSrv.GetLiteral (prop), true);
	                    var propValue = "";
	                    var propType = "";
	                    try {
	                        propValue = new String (cc[prop]);
	                    }
	                    catch (e)
	                    { }
	                   
	                    if (propValue.lastIndexOf ("function") != -1)
	                    {
	                        propValue = "";
	                        propType = "function";
	                    }
	                    
	                    this.mDataSource.Assert (propRes, this.arcValue, this.rdfSrv.GetLiteral (propValue), true);
	                    this.mDataSource.Assert (propRes, this.arcType, this.rdfSrv.GetLiteral (propType), true);
                        this.mDataSource.Assert (propRes, this.arcShow, this.litTrue, true);
                        this.mDataSource.Assert (propRes, this.arcHasNote, this.litNoNote, true);
                        this.mDataSource.Assert (propRes, this.arcHasBookmark, this.litNoBookmark, true);
	                    ifaceCont.AppendElement (propRes);
	                }
	                clCont.AppendElement (ifaceRes);
	            }
	        }
	        this.mDataSource.Assert (res, this.arcLoaded, this.litTrue, true);
	        
            this.__TO_BE_SAVED__ = true;
            // this.saveDataSourceToFile ();
	        this.refreshTree (this._lastView);
	    }
	}, // end gXpcomViewer.loadClassAtResource (aRes)
	
	
	/**
	 * Get details from interface on given resource and put 
	 * these details into properties tree.
	 * 
	 * @param  nsIRDFResource
	 */
	loadInterfaceAtResource : function (aRes)
	{
	    // this method is applicable only on Components.interfaces view
	    if (this._lastView == this.uriIfaces)
	    {
	        var res = this.rdfSrv.GetResource (aRes.Value);
	        var target = this.mDataSource.GetTarget (aRes, this.arcLabel, true);
	        var ifaceName = "";
	   
	        if (target instanceof this.nsIRDFLiteral)
	            ifaceName = target.Value;
	       
	        if (ifaceName != "" && 
	            typeof (Components.interfaces[ifaceName]) != "undefined")
	        {
	            var ifaceCont = Components.classes["@mozilla.org/rdf/container-utils;1"].
                                createInstance (Components.interfaces.nsIRDFContainerUtils).
                                MakeSeq (this.mDataSource, res);
	           
	            for (var ifaceProp in Components.interfaces[ifaceName])
	            {
	                var propRes = this.rdfSrv.GetResource (aRes.Value + ":" + ifaceProp);
	                this.mDataSource.Assert (propRes, this.arcLabel, this.rdfSrv.GetLiteral (ifaceProp), true);
	                this.mDataSource.Assert (propRes, this.arcValue, this.rdfSrv.GetLiteral (Components.interfaces[ifaceName][ifaceProp]), true);
                    this.mDataSource.Assert (propRes, this.arcShow, this.litTrue, true);
                    this.mDataSource.Assert (propRes, this.arcHasNote, this.litNoNote, true);
                    this.mDataSource.Assert (propRes, this.arcHasBookmark, this.litNoBookmark, true);
	                ifaceCont.AppendElement (propRes);
	            }
	        }
	        this.mDataSource.Assert (res, this.arcLoaded, this.litTrue, true);
            this.__TO_BE_SAVED__ = true;
	        this.refreshTree (this._lastView);
	    }
    }, // end gXpcomViewer.loadInterfaceAtResource (aRes)
    
    
    /**
     * Create Components.classes container
     * 
     * @todo    Show progress in statusbar-progress
     */
    createComponentsClasses : function ()
    {
        var classesCont = null;
        try
        {
            // try to check if exist container in our datasource
            classesCont = Components.classes["@mozilla.org/rdf/container;1"].
                createInstance (Components.interfaces.nsIRDFContainer).
                Init (this.mDataSource, this.resClasses);
        }
        catch (e)
        {
            // create Components.results
            classesCont = Components.classes["@mozilla.org/rdf/container-utils;1"].
                createInstance (Components.interfaces.nsIRDFContainerUtils).
                MakeSeq (this.mDataSource, this.resClasses);
        }
        
        for (var cl in Components.classes)
        {
            var newClassRes = this.rdfSrv.GetResource (this.uriClasses + ":" + cl);
           
            this.mDataSource.Assert (newClassRes, this.arcLabel, this.rdfSrv.GetLiteral (cl), true);
            this.mDataSource.Assert (newClassRes, this.arcType, this.litTypeClass, true);
            this.mDataSource.Assert (newClassRes, this.arcShow, this.litTrue, true);
            this.mDataSource.Assert (newClassRes, this.arcHasNote, this.litNoNote, true);
            this.mDataSource.Assert (newClassRes, this.arcHasBookmark, this.litNoBookmark, true);
               
            classesCont.AppendElement (newClassRes);
        }
        
        this.rootCont.AppendElement (this.resClasses);
    },
    
    
    /**
     * Create Components.interfaces container
     * 
     * @todo    Show progress in statusbar-progress
     */
    createComponentsInterfaces : function ()
    {
        var ifacesCont = null;
        try
        {
            ifacesCont = Components.classes["@mozilla.org/rdf/container;1"].
                createInstance (Components.interfaces.nsIRDFContainer).
                Init (this.mDataSource, this.resIfaces);
        }
        catch (e)
        {
            ifacesCont = Components.classes["@mozilla.org/rdf/container-utils;1"].
                createInstance (Components.interfaces.nsIRDFContainerUtils).
                MakeSeq (this.mDataSource, this.resIfaces);
        }
        
        for (var iface in Components.interfaces)
        {
            var newIfaceRes = this.rdfSrv.GetResource (this.uriIfaces + ":" + iface);
           
            this.mDataSource.Assert (newIfaceRes, this.arcLabel, this.rdfSrv.GetLiteral (iface), true);
            this.mDataSource.Assert (newIfaceRes, this.arcType, this.litTypeIface, true);
            this.mDataSource.Assert (newIfaceRes, this.arcShow, this.litTrue, true);
            this.mDataSource.Assert (newIfaceRes, this.arcHasNote, this.litNoNote, true);
            this.mDataSource.Assert (newIfaceRes, this.arcHasBookmark, this.litNoBookmark, true);
               
            ifacesCont.AppendElement (newIfaceRes);
        }
        
        this.rootCont.AppendElement (this.resIfaces);
    },
    
    
    /**
     * Create Components.results container
     * 
     * @todo    Show progress in statusbar-progress
     */
    createComponentsResults : function ()
    {
        var resultsCont = null;
        try
        {
            resultsCont = Components.classes["@mozilla.org/rdf/container;1"].
                createInstance (Components.interfaces.nsIRDFContainer).
                Init (this.mDataSource, this.resResults);
        }
        catch (e)
        {
            resultsCont = Components.classes["@mozilla.org/rdf/container-utils;1"].
                createInstance (Components.interfaces.nsIRDFContainerUtils).
                MakeSeq (this.mDataSource, this.resResults);
        }
        
        for (var result in Components.results)
        {
            var newResultRes = this.rdfSrv.GetResource (this.uriResults + ":" + result);
           
            this.mDataSource.Assert (newResultRes, this.arcLabel, this.rdfSrv.GetLiteral (result), true);
            this.mDataSource.Assert (newResultRes, this.arcValue, this.rdfSrv.GetLiteral (Components.results[result]), true);
            this.mDataSource.Assert (newResultRes, this.arcType, this.litTypeResult, true);
            this.mDataSource.Assert (newResultRes, this.arcShow, this.litTrue, true);
            this.mDataSource.Assert (newResultRes, this.arcHasNote, this.litNoNote, true);
            this.mDataSource.Assert (newResultRes, this.arcHasBookmark, this.litNoBookmark, true);
               
            resultsCont.AppendElement (newResultRes);
        }
        
        this.rootCont.AppendElement (this.resResults);
    }, // end gXpcomViewer.createComponentsResults ()	
	
	
	/**
	 * Create new blank datasource. This means create base container 
	 * and sub-container which is selected by _lastView.
	 */
	createBlankDataSource : function ()
	{
        this.mDataSource = Components.classes[this.inMemoryDS].createInstance (this.nsIRDFDS);
        
        // initialize root container
        this.rootCont = Components.classes["@mozilla.org/rdf/container-utils;1"].
            createInstance (Components.interfaces.nsIRDFContainerUtils).
            MakeSeq (this.mDataSource, this.resBase);
        
        switch (this._lastView)
        {
            case this.uriClasses:
                this.createComponentsClasses ();
                break;
            case this.uriIfaces:
                this.createComponentsInterfaces ();
                break;
            case this.uriResults:
                this.createComponentsResults ();
                break;
        }
	}, // end gXpcomViewer.createBlankDataSource (aType)
	
	
	/**
	 * Initialize local file with stored datasource. Firstly try obtain 
	 * location from custom key and if this fail try default location in 
	 * sapphix working directory.
	 */
	initializeDataSourceFile : function ()
	{
		this.mDataSourceFile = Components.classes["@mozilla.org/file/directory_service;1"].
			getService(Components.interfaces.nsIProperties).
			get("ProfD", Components.interfaces.nsIFile);
		
		this.mDataSourceFile.append ("extensions");
		this.mDataSourceFile.append ("{85CBA252-45E8-4e00-AF4D-E7EE9D462A9B}");
		this.mDataSourceFile.append ("xpcom.rdf");
	}, // end gXpcomViewer.initializeDataSourceFile ()
	
	
	/**
	 * Load datasource of photosets from local file. Before first 
	 * use of this method is needed to call initializeDataSourceFile ().
	 */
	loadDataSourceFromFile : function ()
	{
		var stream1 = Components.classes["@mozilla.org/network/file-input-stream;1"].
			createInstance (Components.interfaces.nsIFileInputStream);
		var stream2 = Components.classes["@mozilla.org/scriptableinputstream;1"].
			createInstance (Components.interfaces.nsIScriptableInputStream);
		var parser = Components.classes["@mozilla.org/rdf/xml-parser;1"].
			createInstance (Components.interfaces.nsIRDFXMLParser);
		
		var rootUri = this.IOSrv.newURI (this.uriBase, null, null);
		var content = new String ("");
		
		this.mDataSource = Components.classes[this.inMemoryDS].
			createInstance (this.nsIRDFDS);
		
		try
		{
			stream1.init (this.mDataSourceFile, 1, 0, false);
			stream2.init (stream1);
			content += stream2.read (-1);
			parser.parseString (this.mDataSource, rootUri, content);
			stream2.close ();
			stream1.close ();
		}
		catch (e)
		{
			Components.utils.reportError ("Error when reading datasource from file!\n" + e);
		}
        
        // initialize root container
        this.rootCont = Components.classes["@mozilla.org/rdf/container;1"].
                createInstance (Components.interfaces.nsIRDFContainer).
                Init (this.mDataSource, this.resBase);
	}, // end gXpcomViewer.loadDataSourceFromFile ()
	
	
	/**
	 * Save datasource to local file.
	 */
	saveDataSourceToFile : function ()
	{
		var outputStream = 
		{
			data : "",
			close : function () {},
			flush : function () {},
			write : function (buffer, count) 
			{
				this.data += buffer;
				return count;
			},
			writeFrom : function (stream, count) {},
			isNonBlocking : false
		}
		
		var serializer = Components.classes["@mozilla.org/rdf/xml-serializer;1"].
			createInstance(Components.interfaces.nsIRDFXMLSerializer);
		var stream = Components.classes["@mozilla.org/network/file-output-stream;1"].
			createInstance (Components.interfaces.nsIFileOutputStream);
		
		try
		{	
			stream.init (this.mDataSourceFile, 0x02 | 0x08 | 0x20, 0664, 0);
			serializer.init (this.mDataSource);
			serializer.QueryInterface (Components.interfaces.nsIRDFXMLSource);
			serializer.Serialize (stream);
			stream.close ();
		}
		catch (e)
		{
			Components.utils.reportError (e);
		}
	}, // end gXpcomViewer.saveDataSourceToFile ()
	
	
	
	/////////////////////////////////////////////
	// Bookmarks and notes methods
	
	
		
	/**
	 * Append note on the selected resource item (component, interface)
	 */
	addNote : function ()
	{
        var resources = this.getSelectedTreeItems ();
        
        if (resources.length == 0)
        {   
            alert (this.mStrBundle.getString ("addNoteError_noSelected"));
            return;
        }
        else 
        {
            for (var res in resources)
            {
                // alert (resources[res]);
                
                var noteDialog = window.
                    open ("chrome://xpcomviewer/content/notedialog.xul", 
                    "xpcomviewer:noteDialog", "chrome,all,modal");
                noteDialog.focus ();
            }
		}
	}, // end gXpcomViewer.addNote ()
	
	
	/**
	 * Append bookmark on the selected resource item (component, interface)
	 */
	addBookmark : function ()
	{
        var resources = this.getSelectedTreeItems ();
        var change = false;
        
        if (resources.length == 0)
        {
            alert (this.mStrBundle.getString ("addBookmarkError_noSelected"));
            return;
        }
        else 
        {
            for (var res in resources)
            {
                // alert (resources[res]);
                
                var hasBookmark = "";
                var target = this.mDataSource.GetTarget (resources[res], 
                    this.arcHasBookmark, true);
                    
                if (target instanceof this.nsIRDFLiteral)
                    hasBookmark = target.Value;
                
                if (hasBookmark == "noBookmark")
                {
                    this.mDataSource.Change (resources[res], 
                        this.arcHasBookmark, this.litNoBookmark, 
                        this.litIsBookmark, true);
                    
                    gBookmarksDataSource.addBookmark (resources[res]);
                    change = true;
                }
            }
            
            if (change)
                this.refreshTree (this._lastView);
		}
	}, // end gXpcomViewer.addBookmark ()
    
    
	
	/////////////////////////////////////////////
	// Bookmarks and notes methods
		
	
	/**
	 * Apply filter on XPCOM datasource and refresh the main 
     * tree. 
     * 
     * @param   string
	 */
	applyFilter : function (aFilterType)
	{
        // our filter describing object
        var filter = { type    : "",  pattern : "" };
        
        // check arguments
        if ((this._lastView.indexOf ("results") != -1) && 
            aFilterType == "implemented_by") 
            return
            
        // for 'with_bookmark' and 'with_note' filter types
        if (aFilterType.indexOf ("with_") == 0) 
            filter.type = aFilterType.replace ("with_", "");
        else 
        {
            // for 'containing', 'starting_with' and 'implemented_by' 
            // filter types
            filter.type = aFilterType;
            var pattern = prompt (this.mStrBundle.getFormattedString ("filterPromptTitle", [aFilterType]));
            
            if (pattern == "")
            {
                alert (this.mStrBundle.getString ("filterPromptError"));
                return;
            }
            else
                filter.pattern = pattern;
        }
        
        this._isFilterApplied = this._applyFilter (filter);
        
        if (this._isFilterApplied)
            this.refreshTree (this._lastView);
	}, // end gXpcomViewer.applyFilter ()
    
    
    /** 
     * Perform filter on datasource. In fact we only add or remove 
     * show property in every resource in container which is currently 
     * displayed according to if if meets filter pattern or not.
	 * 
     * @param   object
     * @return  boolean
     */
    _applyFilter : function (aFilter)
    {
        // 1) initialize container which is currently displayed
        var cont = Components.classes["@mozilla.org/rdf/container;1"].
                   createInstance (Components.interfaces.nsIRDFContainer);
        try {
            cont.Init (this.mDataSource, this.rdfSrv.GetResource (this._lastView));
        } catch (e) {
            return false;
        }
        
        // 2) go through all its elements 
        var elements = cont.GetElements ();
        var changedAny = false;
        
        while (elements.hasMoreElements ())
        {
            var elm = elements.getNext ();
            var result = false;
            
            // 3) find out if resource meets filter patter or not
            if (aFilter.type == "bookmark")
                result = this.mDataSource.hasArcOut (elm, this.arcHasBookmark);
            else if (aFilter.type == "note")
                result = this.mDataSource.hasArcOut (elm, this.arcHasNote);
            else if (aFilter.type == "containing" || 
                     aFilter.type == "starting_with")
            {var label = "";
                var target = this.mDataSource.GetTarget (elm, this.arcLabel, true);
                
                if (target instanceof this.nsIRDFLiteral)
                    label = target.Value;
                
                var _result = false;
                if (label != "")
                    _result = label.indexOf (aFilter.pattern);
                
                if (_result == -1)
                    result = false;
                else 
                {
                    if (aFilter.type == "starting_with")
                        result = (_result == 0) ? true : false;
                    else
                        result = true;
                }
            }
            else if (aFilter.type == "implemented_by")
            {
            
            }
            
            // 4) add/remove show literal on value of result
            if (result)
            {
                if (!this.mDataSource.hasArcOut (elm, this.arcShow))
                {
                    this.mDataSource.Assert (elm, this.arcShow, this.litTrue, true);
                    changedAny = true;
                }
            }
            else
            {
                if (this.mDataSource.hasArcOut (elm, this.arcShow))
                {
                    this.mDataSource.Unassert (elm, this.arcShow, this.litTrue, true);
                    changedAny = true;
                }
            }
        }
        
        return changedAny;
    }, // end gXpcomViewer._applyFilter (aFilter)
    
    /**
     * Remove any applied filter.
     */
    removeFilter : function ()
    {
        // 1) initialize container which is currently displayed
        var cont = Components.classes["@mozilla.org/rdf/container;1"].
                   createInstance (Components.interfaces.nsIRDFContainer);
        try {
            cont.Init (this.mDataSource, this.rdfSrv.GetResource (this._lastView));
        } catch (e) {
            return;
        }
        
        // 2) go through all its elements 
        var elements = cont.GetElements ();
        while (elements.hasMoreElements ())
        {
            var elm = elements.getNext ();
            
            // 3) add show literal on value of result
            if (!this.mDataSource.hasArcOut (elm, this.arcShow))
                this.mDataSource.Assert (elm, this.arcShow, this.litTrue, true);
        }
        
        // 4) refresh tree and
        this._isFilterApplied = false;
        this.refreshTree (this._lastView);
    } // end gXpcomViewer.removeFilter ()
	
} // end of gXpcomViewer ()
